#ifndef _ENVMIRROR_H_
#define _ENVMIRROR_H_

#include <D3dx8.h>
#include <D3d8types.h.>

typedef struct
{
  FLOAT   x,  y,  z;
  FLOAT  nx, ny, nz;
  FLOAT  u, v;
} mirrorVertex;

class CEnvMirror
{
private:
  mirrorVertex m_Vertices[4];
  DWORD m_dwSizeofVertices;
  DWORD m_dwSizeofIndices;
  BOOL m_bTriLinearFilter;
  BOOL m_bBehindMirror;

  LPDIRECT3DVERTEXBUFFER8 m_pVB;
  LPDIRECT3DINDEXBUFFER8 m_pIB;
  LPDIRECT3DTEXTURE8 m_pMirrorTexture;
  LPDIRECT3DSURFACE8 m_pMirrorSurf;
  LPD3DXRENDERTOSURFACE m_pMirrorRenderToSurface;
  UINT m_uiMirrorWidth;
  UINT m_uiMirrorHeight;

  D3DXVECTOR3 m_vecMirrorNormal;    // Front facing normal of mirror plane
  D3DXVECTOR3 m_vecMirrorUp;        // Up vector of mirror
  D3DXVECTOR3 m_vecMirrorTL;        // Vector for Top Left mirror corner
  D3DXVECTOR3 m_vecWorldUp;         // Y-axis for world coordinate system used
  D3DXVECTOR3 m_vecMirrorLTR;       // Left To Right vec for mirror
  D3DXPLANE   m_plnMirror;          // Plane matrix for mirror plane
  D3DXMATRIX  m_matReflect;         // Reflection matrix for mirror plane
  D3DXMATRIX  m_matMirrorRotXY;     // Rotation matrix to rotate mirror onto xy plane
  FLOAT       m_fDistance;          // Perpendicular distance of camera to mirror
  FLOAT       m_fFarClip;           // Far clip value used for 2nd pass reflection rendering. Default to 10000



public:
  CEnvMirror(D3DXVECTOR3 topLeft, D3DXVECTOR3 topRight, D3DXVECTOR3 bottomRight, D3DXVECTOR3 bottomLeft);
  ~CEnvMirror();

public:
  HRESULT     InitDeviceObjects(LPDIRECT3DDEVICE8 m_pd3dDevice);                      
  HRESULT     RestoreDeviceObjects(LPDIRECT3DDEVICE8 m_pd3dDevice);                   
  HRESULT     Render(LPDIRECT3DDEVICE8 m_pd3dDevice);
  HRESULT     InvalidateDeviceObjects();
  HRESULT     DeleteDeviceObjects();

  HRESULT     SetupRenderTarget(LPDIRECT3DDEVICE8 m_pd3dDevice);
  HRESULT     BeginRenderToSurface(LPDIRECT3DDEVICE8 m_pd3dDevice, D3DXMATRIX matPosition);
  void        EndRenderToSurface(LPDIRECT3DDEVICE8 m_pd3dDevice);
  BOOL        IsCamBehindMirror();
  D3DXPLANE   GetMirrorPlane(D3DXPLANE& plane);
  FLOAT       GetPerpendicularDistance();
  void        SetFarClip(FLOAT value);

};


__inline float Max4f(float a, float b, float c, float d)
{
  float r;
  if (a > b)
  {
    if (a > c)
      (a > d)?(r = a):(r = d);
    else
      (c > d)?(r = c):(r = d);
  }
  else
  {
    if (b > c)
      (b > d)?(r = b):(r = d);
    else
      (c > d)?(r = c):(r = d);
  }
  return r;
}
__inline float Min4f(float a, float b, float c, float d)
{
  float r;
  if (a < b)
  {
    if (a < c)
      (a < d)?(r = a):(r = d);
    else
      (c < d)?(r = c):(r = d);
  }
  else
  {
    if (b < c)
      (b < d)?(r = b):(r = d);
    else
      (c < d)?(r = c):(r = d);
  }
  return r;
}

// A vec3 multiplication to be used when testing the direction of the mirror's rotated mirror
// Why isn't there a D3DVec3Multiply???
__inline float Vec3Multiply(const D3DXVECTOR3* v1, const D3DXVECTOR3* v2)
{
  float ans;
  ans = (v1->x*v2->x)+(v1->y*v2->y)+(v1->z*v2->z);
  return ans;
}

#endif _ENVMIRROR_H_