#include "EnvMirror.h"
#include <d3dutil.h>
#include <tchar.h>
#include <d3dapp.h>
#include <dxutil.h>
#include <D3dx8core.h>
#include <math.h>

#define D3DFVF_MIRROR (D3DFVF_XYZ|D3DFVF_NORMAL|D3DFVF_TEX1)
#define PI 3.1415926538
  
// Indices
WORD  g_dwIndices[] =   {0, 1, 2, 0, 2, 3};
D3DXMATRIX g_matMirrorOnXY; // Stores the vec of all four rotated mirror corners. Mirror is rotated to XY plane

CEnvMirror::CEnvMirror(D3DXVECTOR3 topLeft, D3DXVECTOR3 topRight, D3DXVECTOR3 bottomRight, D3DXVECTOR3 bottomLeft)
{
  // Top Left
  m_Vertices[0].x  = topLeft.x;
  m_Vertices[0].y  = topLeft.y;
  m_Vertices[0].z  = topLeft.z;
  m_Vertices[0].u  = 1.0f;
  m_Vertices[0].v  = 0.0f;
  // Top Right
  m_Vertices[1].x = topRight.x;
  m_Vertices[1].y = topRight.y;
  m_Vertices[1].z = topRight.z;
  m_Vertices[1].u  = 0.0f;
  m_Vertices[1].v  = 0.0f;
  // Bottom Right
  m_Vertices[2].x = bottomRight.x;
  m_Vertices[2].y = bottomRight.y;
  m_Vertices[2].z = bottomRight.z;
  m_Vertices[2].u  = 0.0f;
  m_Vertices[2].v  = 1.0f;
  // Bottom Left
  m_Vertices[3].x = bottomLeft.x;
  m_Vertices[3].y = bottomLeft.y;
  m_Vertices[3].z = bottomLeft.z;
  m_Vertices[3].u  = 1.0f;
  m_Vertices[3].v  = 1.0f;

  m_vecWorldUp    = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
  m_vecMirrorTL   = D3DXVECTOR3(m_Vertices[0].x, m_Vertices[0].y, m_Vertices[0].z); // Vector of TL mirror corner
  
  D3DXVECTOR3 vecMirrorDown;
  m_vecMirrorUp   = D3DXVECTOR3(m_Vertices[0].x-m_Vertices[3].x, m_Vertices[0].y-m_Vertices[3].y, m_Vertices[0].z-m_Vertices[3].z);
  vecMirrorDown   = D3DXVECTOR3(m_Vertices[3].x-m_Vertices[0].x, m_Vertices[3].y-m_Vertices[0].y, m_Vertices[3].z-m_Vertices[0].z);
  m_vecMirrorLTR  = D3DXVECTOR3(m_Vertices[3].x-m_Vertices[2].x, m_Vertices[3].y-m_Vertices[2].y, m_Vertices[3].z-m_Vertices[2].z);
  D3DXVec3Normalize(&m_vecMirrorLTR, &m_vecMirrorLTR);

  // Calculate with DOWN and LeftToRight vector. This is bcos of LH convention used
  D3DXVec3Cross(&m_vecMirrorNormal, &vecMirrorDown, &m_vecMirrorLTR);
  D3DXVec3Normalize(&m_vecMirrorNormal, &m_vecMirrorNormal);
  D3DXVec3Normalize(&m_vecMirrorUp, &m_vecMirrorUp);


  // Must rotate mirror to the xy plane here.
  // Rotation matrix need to be applied to the virtual cam position every frame to get the correct translation!
  double dRotX, dRotY;
  D3DXVECTOR3 vecXAxis, vecYAxis;
  D3DXMATRIX matTemp;
  
  D3DXMatrixIdentity(&m_matMirrorRotXY);
  D3DXMatrixIdentity(&matTemp);
  vecXAxis = D3DXVECTOR3(1.0f, 0.0f, 0.0f);
  vecYAxis = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
  dRotX = D3DXVec3Dot(&m_vecMirrorLTR, &vecXAxis);
  dRotY = D3DXVec3Dot(&m_vecMirrorUp, &vecYAxis);
  dRotX = acos(dRotX);
  dRotY = acos(dRotY);
  D3DXMatrixRotationY(&m_matMirrorRotXY, ((float)dRotX)); // Rotate about Y
  D3DXMatrixRotationX(&matTemp, ((float)dRotY));          // Rotate about X
  D3DXMatrixMultiply(&m_matMirrorRotXY, &m_matMirrorRotXY, &matTemp);

  // We need to test whether the mirror will be facing the +ve z direction or not
  // This is very impt as any projection matrix calculation will only work when the camera is look at +ve z direction
  // We need to think as the VIRTUAL CAMERA!! The virtual camera must be looking at the +ve z direction!
  // We can find out whether the virtual camera is looking at +ve by doing the following
  //   - Take the mirror calculated and rotate it using the m_matMirrorRotXY matrix
  //   - If the rotated normal points towards +ve then we do not need to do anything, else we must compute the
  //   - rotation matrix again by negating the rotation angle about the Y axis (dRotX)
  //   - I am assuming default clockwise rotation here. Negating the angle rotates the mirror in counter clockwise
  //   - Note: Mirror's normal is actually the direction of the virtual camera
  D3DXVECTOR3 ThirdColOfRotXY = D3DXVECTOR3(m_matMirrorRotXY._13, m_matMirrorRotXY._23, m_matMirrorRotXY._33);
  // Taking short cut here. We need to test only the z value of the rotated normal, so its just a vec3 * vec3
  if (0.0f > Vec3Multiply(&m_vecMirrorNormal, &ThirdColOfRotXY)) // Need to recompute!
  {
    D3DXMatrixRotationY(&m_matMirrorRotXY, ((float)(-dRotX))); // Rotate CCW !
    D3DXMatrixMultiply(&m_matMirrorRotXY, &m_matMirrorRotXY, &matTemp);
  }

  g_matMirrorOnXY._11  = m_Vertices[0].x;
  g_matMirrorOnXY._12  = m_Vertices[0].y;
  g_matMirrorOnXY._13  = m_Vertices[0].z;
  g_matMirrorOnXY._21  = m_Vertices[1].x;
  g_matMirrorOnXY._22  = m_Vertices[1].y;
  g_matMirrorOnXY._23  = m_Vertices[1].z;
  g_matMirrorOnXY._31  = m_Vertices[2].x;
  g_matMirrorOnXY._32  = m_Vertices[2].y;
  g_matMirrorOnXY._33  = m_Vertices[2].z;
  g_matMirrorOnXY._41  = m_Vertices[3].x;
  g_matMirrorOnXY._42  = m_Vertices[3].y;
  g_matMirrorOnXY._43  = m_Vertices[3].z;
  D3DXMatrixMultiply( &g_matMirrorOnXY, &g_matMirrorOnXY, &m_matMirrorRotXY); // Rotate to xy plane
  // Set to 1 so that translation will be taken into account during mat multiplication
  g_matMirrorOnXY._14  = 1.0f; g_matMirrorOnXY._24  = 1.0f; g_matMirrorOnXY._34  = 1.0f; g_matMirrorOnXY._44  = 1.0f;


  // Set all the normals
  m_Vertices[0].nx = m_vecMirrorNormal.x;
  m_Vertices[0].ny = m_vecMirrorNormal.y;
  m_Vertices[0].nz = m_vecMirrorNormal.z;
  m_Vertices[1].nx = m_vecMirrorNormal.x;
  m_Vertices[1].ny = m_vecMirrorNormal.y;
  m_Vertices[1].nz = m_vecMirrorNormal.z;
  m_Vertices[2].nx = m_vecMirrorNormal.x;
  m_Vertices[2].ny = m_vecMirrorNormal.y;
  m_Vertices[2].nz = m_vecMirrorNormal.z;
  m_Vertices[3].nx = m_vecMirrorNormal.x;
  m_Vertices[3].ny = m_vecMirrorNormal.y;
  m_Vertices[3].nz = m_vecMirrorNormal.z;

  m_dwSizeofVertices  = sizeof(m_Vertices);
  m_dwSizeofIndices   = sizeof(g_dwIndices);
  m_pVB               = NULL;
  m_pIB               = NULL;
  m_uiMirrorWidth     = 512;
  m_uiMirrorHeight    = 512;
  m_fDistance         = 0.0f;
  m_fFarClip          = 10000.0f;

  m_bBehindMirror     = FALSE;

  ZeroMemory(&m_plnMirror, sizeof(m_plnMirror));
  D3DXPlaneFromPointNormal(&m_plnMirror, &topLeft, &m_vecMirrorNormal);

  D3DXMatrixReflect(&m_matReflect, &m_plnMirror);
}

CEnvMirror::~CEnvMirror(){}


HRESULT CEnvMirror::InitDeviceObjects(LPDIRECT3DDEVICE8 m_pd3dDevice)
{
  return S_OK;
}

HRESULT CEnvMirror::RestoreDeviceObjects(LPDIRECT3DDEVICE8 m_pd3dDevice)
{
  // Setup our render target
  if (FAILED(SetupRenderTarget(m_pd3dDevice)))
    return E_FAIL;

    // Create the vertex buffer
  if( FAILED( m_pd3dDevice->CreateVertexBuffer( m_dwSizeofVertices, 0, D3DFVF_MIRROR, D3DPOOL_MANAGED, &m_pVB ) ) )
        return E_FAIL;

  VOID* pVertices;
  if( FAILED( m_pVB->Lock( 0, m_dwSizeofVertices, (BYTE**)&pVertices, 0 ) ) )
    return E_FAIL;
  
  memcpy( pVertices, m_Vertices, m_dwSizeofVertices);
  m_pVB->Unlock();

	// Create the index buffer
	if( FAILED( m_pd3dDevice->CreateIndexBuffer( m_dwSizeofIndices, 0, D3DFMT_INDEX16, D3DPOOL_MANAGED, &m_pIB ) ) )
		return E_FAIL;

	VOID* pIndices;
	if( FAILED( m_pIB->Lock( 0, m_dwSizeofIndices, (BYTE**)&pIndices, 0 ) ) )
	  return E_FAIL;
	memcpy( pIndices, g_dwIndices, sizeof(g_dwIndices) );
	m_pIB->Unlock();

  return S_OK;
}


HRESULT CEnvMirror::Render(LPDIRECT3DDEVICE8 m_pd3dDevice)
{

  m_pd3dDevice->SetTexture(0, m_pMirrorTexture );
	m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1 );

  m_pd3dDevice->SetStreamSource(0, m_pVB, sizeof(mirrorVertex));
  m_pd3dDevice->SetVertexShader(D3DFVF_MIRROR);
  m_pd3dDevice->SetIndices(m_pIB, 0);
  m_pd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 4, 0, 2);

  return S_OK;
}


HRESULT CEnvMirror::InvalidateDeviceObjects()
{
  SAFE_RELEASE(m_pVB);
  SAFE_RELEASE(m_pIB);

  SAFE_RELEASE(m_pMirrorRenderToSurface);
  SAFE_RELEASE(m_pMirrorSurf);
  SAFE_RELEASE(m_pMirrorTexture);

  return S_OK;
}

HRESULT CEnvMirror::DeleteDeviceObjects()
{
  return S_OK;
}


HRESULT CEnvMirror::SetupRenderTarget(LPDIRECT3DDEVICE8 m_pd3dDevice)
{
  
  HRESULT hr;

  D3DDISPLAYMODE mode;
  m_pd3dDevice->GetDisplayMode(&mode);

  // Create mirror texture
  if(FAILED(hr=D3DXCreateTexture(m_pd3dDevice, m_uiMirrorWidth, m_uiMirrorHeight, 1, D3DUSAGE_RENDERTARGET, mode.Format, D3DPOOL_DEFAULT, &m_pMirrorTexture)))
  {
    return hr;
  }

  D3DSURFACE_DESC desc;
  hr = m_pMirrorTexture->GetSurfaceLevel(0, &m_pMirrorSurf);
  m_pMirrorSurf->GetDesc(&desc);

  // Set up render target
  if(FAILED(hr=D3DXCreateRenderToSurface(m_pd3dDevice, desc.Width, desc.Height, mode.Format, TRUE, D3DFMT_D16, &m_pMirrorRenderToSurface)))
  {
    return hr;
  }

  return S_OK;
}


HRESULT CEnvMirror::BeginRenderToSurface(LPDIRECT3DDEVICE8 m_pd3dDevice, D3DXMATRIX matPosition)
{
  // Calculate perpendicular distance from cam to mirror plane
  D3DXVECTOR3 vecCamPos(matPosition._41, matPosition._42, matPosition._43);  // Actual virtual cam position
  D3DXVECTOR3 vecMirrorToCam = vecCamPos-m_vecMirrorTL;
  m_fDistance = D3DXVec3Dot( &vecMirrorToCam, &m_vecMirrorNormal);
  if (m_fDistance<=0.0f) // Exit immediately
  {
    m_bBehindMirror = TRUE;
    return S_OK;
  }
  else
    m_bBehindMirror = FALSE;
  
  D3DXMATRIX mat;
  D3DXMatrixTranslation ( &mat, matPosition._41, matPosition._42, matPosition._43 );  // Apply cam's translation
  D3DXMatrixMultiply( &mat, &mat, &m_matReflect); // Virtual pos
  
  // We use the current virtual cam pos to translate the mirror in cam space so that the cam is look at +ve z at the origin
  // The virtual cam pos is rotated using the pre-computed m_matMirrorXYRot which is used to rotate the mirror from world to cam space 
  // The mirror coor in cam space needs to be translated by the -ve value of the rotated virtual cam pos.
  D3DXMATRIX matVirtualCamTrans;
  D3DXMatrixTranslation( &matVirtualCamTrans, mat._41, mat._42, mat._43);  // Translate back to origin
  D3DXMatrixMultiply(&matVirtualCamTrans, &matVirtualCamTrans, &m_matMirrorRotXY);  // Rotate virtual pos
  // We need to make a new translation so that the rotation info in the matrix is not taken into effect when computing the view matrix
  // We just want the rotated virtual cam position and nothing else.
  D3DXMatrixTranslation( &matVirtualCamTrans, -matVirtualCamTrans._41, -matVirtualCamTrans._42, -matVirtualCamTrans._43);

  D3DXVECTOR3 vecVirtualCamPos(mat._41, mat._42, mat._43);  // Actual virtual cam position
  D3DXMATRIX matMirrorProject;
  D3DXMATRIX matMirrorView;
  D3DXMatrixLookAtLH(&matMirrorView, &vecVirtualCamPos, &vecCamPos, &m_vecWorldUp); // Virtual Cam must never rotate
  D3DXMATRIX matMirrorInCamSpace;
  D3DXMatrixMultiply( &matMirrorInCamSpace, &g_matMirrorOnXY, &matVirtualCamTrans);

  // Calculate the projection matrix for the virtual cam
  D3DXMatrixPerspectiveOffCenterLH(&matMirrorProject, 
                                    Min4f(matMirrorInCamSpace._11, matMirrorInCamSpace._21, matMirrorInCamSpace._31, matMirrorInCamSpace._41),
                                    Max4f(matMirrorInCamSpace._11, matMirrorInCamSpace._21, matMirrorInCamSpace._31, matMirrorInCamSpace._41),
                                    Min4f(matMirrorInCamSpace._12, matMirrorInCamSpace._22, matMirrorInCamSpace._32, matMirrorInCamSpace._42),
                                    Max4f(matMirrorInCamSpace._12, matMirrorInCamSpace._22, matMirrorInCamSpace._32, matMirrorInCamSpace._42),
                                    m_fDistance, m_fFarClip);

  m_pd3dDevice->SetTransform(D3DTS_PROJECTION, &matMirrorProject);
  m_pd3dDevice->SetTransform(D3DTS_VIEW, &matMirrorView);

  return m_pMirrorRenderToSurface->BeginScene(m_pMirrorSurf, NULL);
}

void CEnvMirror::EndRenderToSurface(LPDIRECT3DDEVICE8 m_pd3dDevice)
{
  m_pMirrorRenderToSurface->EndScene();
}

BOOL CEnvMirror::IsCamBehindMirror()
{
  return m_bBehindMirror;
}

D3DXPLANE CEnvMirror::GetMirrorPlane(D3DXPLANE& plane)
{
  plane = m_plnMirror;
  return m_plnMirror;
}

FLOAT CEnvMirror::GetPerpendicularDistance()
{
  return m_fDistance;
}

void CEnvMirror::SetFarClip(FLOAT value)
{
  if (value>0.0f)
    m_fFarClip = value;
}