#define STRICT
#include <tchar.h>
#include <math.h>
#include <stdio.h>
#include <D3DX8.h>
#include "D3DApp.h"
#include "D3DFile.h"
#include "D3DFont.h"
#include "D3DUtil.h"
#include "DXUtil.h"

#include "EnvMirror.h"

enum FillModeEnum{
  POINT_MODE  = 0,
  SHADE_MODE  = 1,
  WIRE_MODE   = 2
};


class CMyD3DApplication : public CD3DApplication
{
  BOOL m_bKeyInfo;
  D3DXMATRIX m_matProject;
  D3DXMATRIX m_matView;
  D3DXMATRIX m_matWorld;
  D3DXMATRIX m_matPosition;
  D3DXMATRIX m_matSphere;
  D3DXMATRIX m_matVC;
  D3DXMATRIX m_matReflect;

  D3DXPLANE m_plnMirror;

  CD3DFont* m_pFont;
  CD3DMesh* m_pSkyBox;
  CD3DMesh* m_pSphere;

  BYTE m_bKey[256];

  // Fill mode
  FillModeEnum FillMode;

  // Variables for determining view position
	D3DXVECTOR3	m_pvVelocity;
  D3DXVECTOR3 m_pvAngularVelocity;
  D3DXVECTOR3	m_pvPosition;
	D3DXVECTOR3	m_pvYPR;

	FLOAT m_fAngularSpeed;
  FLOAT m_fSpeed;

  // Environment reflection mirror
  CEnvMirror *m_pMirror;

protected:
  HRESULT RenderScene();
  HRESULT RenderMirror( CONST D3DXMATRIX *pView, CONST D3DXMATRIX *pProject );

  HRESULT ConfirmDevice( D3DCAPS8*, DWORD, D3DFORMAT );
  HRESULT OneTimeSceneInit();
  HRESULT InitDeviceObjects();
  HRESULT RestoreDeviceObjects();
  HRESULT InvalidateDeviceObjects();
  HRESULT DeleteDeviceObjects();
  HRESULT Render();
  HRESULT FrameMove();
  HRESULT FinalCleanup();

public:
  LRESULT MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

  CMyD3DApplication();
};


INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
    CMyD3DApplication d3dApp;

    if( FAILED( d3dApp.Create( hInst ) ) )
        return 0;

    return d3dApp.Run();
}



CMyD3DApplication::CMyD3DApplication()
{
  m_strWindowTitle    = _T("Virtual Camera Position Reflections");
  m_bUseDepthBuffer   = TRUE;
  m_bKeyInfo          = FALSE;
  
  FillMode            = SHADE_MODE;

  m_pFont             = NULL;
  m_pSkyBox           = NULL;
  m_pSphere           = NULL;

  m_fAngularSpeed     = 0.5f;
  m_fSpeed            = 1.0f;

  m_pvAngularVelocity = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
  m_pvVelocity        = D3DXVECTOR3(0.0f, 0.0f, 0.0f);

  m_dwCreationWidth   = 600;
  m_dwCreationHeight  = 400;

  ZeroMemory(m_bKey, 256);
}



HRESULT CMyD3DApplication::OneTimeSceneInit()
{
  D3DXMatrixIdentity( &m_matWorld );
  
  m_pFont        = new CD3DFont( _T("Arial"), 10, D3DFONT_BOLD );
  m_pSkyBox      = new CD3DMesh();
  m_pSphere      = new CD3DMesh();


  D3DXVECTOR3 mirrorTL = D3DXVECTOR3( 3.95f, 7.95f, -8.0f);  // Top Left
  D3DXVECTOR3 mirrorTR = D3DXVECTOR3(-3.95f, 7.95f, -8.0f);  // Top Right
  D3DXVECTOR3 mirrorBR = D3DXVECTOR3(-3.95f, 0.0f, -8.0f);  // Bottom Right
  D3DXVECTOR3 mirrorBL = D3DXVECTOR3( 3.95f, 0.0f, -8.0f);  // Bottom Left

  m_pMirror = new CEnvMirror(mirrorTL, mirrorTR, mirrorBR, mirrorBL);

  m_pMirror->GetMirrorPlane(m_plnMirror);
  D3DXMatrixReflect(&m_matReflect, &m_plnMirror);

  if( !m_pFont || !m_pSkyBox || !m_pSphere)
      return E_OUTOFMEMORY;

  return S_OK;
}



HRESULT CMyD3DApplication::FrameMove()
{
  // Animate sphere
  
  D3DXMATRIX mat;
  D3DXMatrixScaling( &m_matSphere, 4.0f, 4.0f, 4.0f );
  D3DXMatrixTranslation ( &mat, 0.0f, 4.0f, 0.0f );
  D3DXMatrixMultiply( &m_matSphere, &m_matSphere, &mat );
  D3DXMatrixRotationX( &mat, -0.9f*m_fTime );
  D3DXMatrixMultiply( &m_matSphere, &m_matSphere, &mat );
  D3DXMatrixRotationY( &mat, 1.055f*m_fTime );
  D3DXMatrixMultiply( &m_matSphere, &m_matSphere, &mat );
  D3DXMatrixTranslation ( &mat, 0.0f, 5.5f, 0.0f );
  D3DXMatrixMultiply( &m_matSphere, &m_matSphere, &mat );
  
  // move/rotate speed ~should be fps independent
  const cTimeScale = 3;

	//************************************************************
	// camera stuff
	//************************************************************
  D3DXVECTOR3 vTrans(0.0f, 0.0f, 0.0f);
  D3DXVECTOR3 vRot(0.0f, 0.0f, 0.0f);

	// Process keyboard input
  if (m_bKey[VK_NUMPAD8])		    vTrans.z += 0.2f;   // Move Forward
  if (m_bKey['W'])      		    vTrans.z += 0.2f;   // Move Forward
  if (m_bKey[VK_NUMPAD2])			  vTrans.z -= 0.2f;   // Move Backward
  if (m_bKey['S'])      			  vTrans.z -= 0.2f;   // Move Backward
  if (m_bKey[VK_NUMPAD7])		    vTrans.x -= 0.15f;  // Slide Left
  if (m_bKey['Q'])      		    vTrans.x -= 0.15f;  // Slide Left
  if (m_bKey[VK_NUMPAD9])		    vTrans.x += 0.15f;  // Slide Right
  if (m_bKey['E'])      		    vTrans.x += 0.15f;  // Slide Right
  if (m_bKey[VK_ADD])           vTrans.y -= 0.1f;   // Slide Down
  if (m_bKey['F'])              vTrans.y -= 0.1f;   // Slide Down
  if (m_bKey[VK_SUBTRACT])      vTrans.y += 0.1f;   // Slide Up
  if (m_bKey['R'])              vTrans.y += 0.1f;   // Slide Up

	if (m_bKey[VK_DIVIDE])			  vRot.y -= 0.1f;     // Pitch Up
	if (m_bKey['T'])	      		  vRot.y -= 0.1f;     // Pitch Up
  if (m_bKey[VK_MULTIPLY])	  	vRot.y += 0.1f;     // Pitch Down
  if (m_bKey['G'])        	  	vRot.y += 0.1f;     // Pitch Down
  if (m_bKey[VK_NUMPAD6])	      vRot.x += 0.1f;     // Turn right
  if (m_bKey['D'])	            vRot.x += 0.1f;     // Turn right
  if (m_bKey[VK_NUMPAD4]) 	    vRot.x -= 0.1f;     // Turn left
  if (m_bKey['A']) 	            vRot.x -= 0.1f;     // Turn left
  if (m_bKey[VK_NUMPAD1])		    vRot.z += 0.02f;    // Rotate Left
  if (m_bKey['Z'])      		    vRot.z += 0.02f;    // Rotate Left
  if (m_bKey[VK_NUMPAD3])		    vRot.z -= 0.02f;    // Rotate Right
  if (m_bKey['C'])      		    vRot.z -= 0.02f;    // Rotate Right
	
	// turn cfSmooth to 0.98f
	const FLOAT cfSmooth = 0.9f;

	// transform and rotation velocity
  m_pvVelocity = m_pvVelocity * cfSmooth + vTrans;
  m_pvAngularVelocity = m_pvAngularVelocity * cfSmooth + vRot;

	// transform and rotation value
  vTrans = m_pvVelocity * m_fElapsedTime * m_fSpeed * cTimeScale;
  vRot = m_pvAngularVelocity* m_fElapsedTime * m_fAngularSpeed * cTimeScale;

  // Update position and view matricies
  D3DXMATRIX matT, matR;
  D3DXQUATERNION qR;

	D3DXMatrixTranslation (&matT, vTrans.x, vTrans.y, vTrans.z);      // step 1
	D3DXMatrixMultiply (&m_matPosition, &matT, &m_matPosition);
	D3DXQuaternionRotationYawPitchRoll (&qR, vRot.x, vRot.y, vRot.z);	// step 2
	D3DXMatrixRotationQuaternion (&matR, &qR);							          // step 3
	D3DXMatrixMultiply (&m_matPosition, &matR, &m_matPosition);			  // step 4
	D3DXMatrixInverse (&m_matView, NULL, &m_matPosition);				        // step 5

  // ********************** CAMERA MOVEMENT END

  return S_OK;
}


HRESULT CMyD3DApplication::RenderScene()
{
  // Render the Skybox
  // Trilinear filtering
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MIPFILTER, D3DTEXF_LINEAR );
  m_pSkyBox->Render(m_pd3dDevice);

  m_pd3dDevice->SetTransform( D3DTS_WORLD, &m_matSphere );
  m_pSphere->Render(m_pd3dDevice); 

  // Restore world mat
  D3DXMATRIX matWorld;
  D3DXMatrixScaling( &matWorld, 1.0f, 1.0f, 1.0f );
  m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld);

  D3DXMatrixTranslation ( &m_matVC, m_matPosition._41, m_matPosition._42, m_matPosition._43 );  // Apply cam's translation
  D3DXMatrixMultiply( &m_matVC, &m_matVC, &m_matReflect); // Virtual pos

  return S_OK;
}


// Renders the mirror
HRESULT CMyD3DApplication::RenderMirror( CONST D3DXMATRIX *pView, CONST D3DXMATRIX *pProject )
{
  D3DXMATRIX matWorld;
  D3DXMatrixScaling( &matWorld, 1.0f, 1.0f, 1.0f );
  D3DXMATRIX matView(*pView);    

  m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );
  m_pd3dDevice->SetTransform( D3DTS_VIEW, &matView );
  m_pd3dDevice->SetTransform( D3DTS_PROJECTION, pProject );

  m_pMirror->Render(m_pd3dDevice);

  return S_OK;
}


HRESULT CMyD3DApplication::Render()
{
  // Render the reflection texture first
  HRESULT hr;
  if (SUCCEEDED(hr = m_pMirror->BeginRenderToSurface(m_pd3dDevice, m_matPosition)))
  {

    if (!(m_pMirror->IsCamBehindMirror()))
    {
      m_pd3dDevice->Clear(0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, 0x00000000, 1.0f, 0L);
      // Render scene to texture
      RenderScene();
      m_pMirror->EndRenderToSurface(m_pd3dDevice);
    }
  }
  else
    return hr;

  // Reset camera view and projection matrix
  m_pd3dDevice->SetTransform( D3DTS_VIEW, &m_matView );
  m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &m_matProject );

  // Clear the backbuffer
  m_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, 0x00000000, 1.0f, 0L );
  // Begin the scene
  if( SUCCEEDED( m_pd3dDevice->BeginScene() ) )
  {
    // Render the scene
    RenderScene();

    // Render the mirror
    RenderMirror( &m_matView, &m_matProject );

    // Output text
    char str[128];
    if (m_bKeyInfo)
    {
      sprintf(str, "----Controls---- (F1 to toggle help)");
      m_pFont->DrawText( 2, 0, D3DCOLOR_ARGB(255,255,255,0), str );
      sprintf(str, "Forward -------- W, Numpad_8");
      m_pFont->DrawText( 2, 15, D3DCOLOR_ARGB(255,255,255,0), str );
      sprintf(str, "Backward ------ S, Numpad_5");
      m_pFont->DrawText( 2, 30, D3DCOLOR_ARGB(255,255,255,0), str );
      sprintf(str, "LeftTurn -------- A, Numpad_4");
      m_pFont->DrawText( 2, 45, D3DCOLOR_ARGB(255,255,255,0), str );
      sprintf(str, "RightTurn ------ D, Numpad_6");
      m_pFont->DrawText( 2, 60, D3DCOLOR_ARGB(255,255,255,0), str );
      sprintf(str, "Up ----------------- R, Subtract");
      m_pFont->DrawText( 2, 75, D3DCOLOR_ARGB(255,255,255,0), str );
      sprintf(str, "Down ------------- F, Add");
      m_pFont->DrawText( 2, 90, D3DCOLOR_ARGB(255,255,255,0), str );
      sprintf(str, "LeftStrafe ------ Q, Numpad_7");
      m_pFont->DrawText( 2, 105, D3DCOLOR_ARGB(255,255,255,0), str );
      sprintf(str, "RightStrafe ---- E, Numpad_9");
      m_pFont->DrawText( 2, 120, D3DCOLOR_ARGB(255,255,255,0), str );
      sprintf(str, "YawLeft --------- Z, Numpad_1");
      m_pFont->DrawText( 2, 135, D3DCOLOR_ARGB(255,255,255,0), str );
      sprintf(str, "YawRight ------ C, Numpad_3");
      m_pFont->DrawText( 2, 150, D3DCOLOR_ARGB(255,255,255,0), str );
      sprintf(str, "PitchUp --------- T, Divide");
      m_pFont->DrawText( 2, 165, D3DCOLOR_ARGB(255,255,255,0), str );
      sprintf(str, "PitchDown ---- G, Multiply");
      m_pFont->DrawText( 2, 180, D3DCOLOR_ARGB(255,255,255,0), str );
      sprintf(str, "Cycle mode -- M");
      m_pFont->DrawText( 2, 195, D3DCOLOR_ARGB(255,255,255,0), str );
    }
    else
    {
      sprintf(str, "%s (F1 to toggle help)", m_strDeviceStats);
      m_pFont->DrawText( 2, 0, D3DCOLOR_ARGB(255,255,255,0), str );
      m_pFont->DrawText( 2, 15, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats );
      sprintf(str, "Camera        : %f, %f, %f", m_matPosition._41, m_matPosition._42, m_matPosition._43);
      m_pFont->DrawText( 2, 30, D3DCOLOR_ARGB(255,255,255,0), str);
      sprintf(str, "Virtual Cam : %f, %f, %f  (Distance : %f)", m_matVC._41, m_matVC._42, m_matVC._43, m_pMirror->GetPerpendicularDistance());
      m_pFont->DrawText( 2, 45, D3DCOLOR_ARGB(255,255,255,0), str);
    }

    // End the scene.
    m_pd3dDevice->EndScene();
  }

  return S_OK;
}


HRESULT CMyD3DApplication::InitDeviceObjects()
{
    // Load the file objects
  if( FAILED( m_pSkyBox->Create( m_pd3dDevice, _T("data/"), _T("TheRoom.x"), FALSE ) ) )
    return D3DAPPERR_MEDIANOTFOUND;

  if( FAILED( m_pSphere->Create( m_pd3dDevice, _T("data/"), _T("sphere.x"), FALSE ) ) )
    return D3DAPPERR_MEDIANOTFOUND;

  // Restore the device-dependent objects
  m_pFont->InitDeviceObjects(m_pd3dDevice);

  m_pMirror->InitDeviceObjects(m_pd3dDevice);

  return S_OK;
}


HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
  // InitDeviceObjects for file objects (build textures and vertex buffers)
  m_pSkyBox->RestoreDeviceObjects(m_pd3dDevice);
  m_pSphere->RestoreDeviceObjects(m_pd3dDevice);
  m_pFont->RestoreDeviceObjects();
  m_pMirror->RestoreDeviceObjects(m_pd3dDevice);

  // Set the transform matrices
  FLOAT fAspect = (FLOAT) m_d3dsdBackBuffer.Width / (FLOAT) m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH( &m_matProject, D3DX_PI * 0.4f, fAspect, 0.5f, 50000.0f );

  // Set view and position matrix
  D3DXMATRIX matRotY;
  D3DXMatrixRotationY(&matRotY, D3DX_PI);
  D3DXMatrixTranslation(&m_matView, 0.0f, 0.0f, 0.0f);
  D3DXMatrixTranslation(&m_matPosition, 0.0f, 3.0f, 0.0f);
  D3DXMatrixMultiply(&m_matPosition, &matRotY, &m_matPosition);

  // Enable bright ambient light
  m_pd3dDevice->SetRenderState( D3DRS_AMBIENT, 0x00ffffff );

  // Enable z-buffer
  m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);


  return S_OK;
}


HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
  m_pSkyBox->InvalidateDeviceObjects();
  m_pSphere->InvalidateDeviceObjects();
  m_pFont->InvalidateDeviceObjects();
  m_pMirror->InvalidateDeviceObjects();

  return S_OK;
}


HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
  m_pFont->DeleteDeviceObjects();
  m_pSkyBox->Destroy();
  m_pSphere->Destroy();
  m_pMirror->DeleteDeviceObjects();

  return S_OK;
}


HRESULT CMyD3DApplication::FinalCleanup()
{
  SAFE_DELETE( m_pFont );
  SAFE_DELETE( m_pSkyBox );
  SAFE_DELETE( m_pSphere );
  delete m_pMirror;

  return S_OK;
}


HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{
  if( dwBehavior & D3DCREATE_PUREDEVICE )
      return E_FAIL;

  return S_OK;
}


LRESULT CMyD3DApplication::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam,
                                    LPARAM lParam )
{
  if (WM_KEYDOWN == uMsg)
  {
    m_bKey[wParam] = 1;
  }
  if (WM_KEYUP == uMsg)
  {
    m_bKey[wParam] = 0;
    if (wParam==VK_F1)
      m_bKeyInfo = !m_bKeyInfo;
    if (wParam=='M')
    {
      switch (FillMode)
      {
      case SHADE_MODE:
        FillMode = WIRE_MODE;
        break;
      case WIRE_MODE:
        FillMode = POINT_MODE;
        break;
      case POINT_MODE:
        FillMode = SHADE_MODE;
      }
    }
    switch (FillMode)
    {
    case SHADE_MODE:
      m_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
      break;
    case WIRE_MODE:
      m_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
      break;
    case POINT_MODE:
      m_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_POINT );
    }
  }

  // Pass remaining messages to default handler
  return CD3DApplication::MsgProc( hWnd, uMsg, wParam, lParam );
}
